import json

from django.contrib import admin, messages
from django.contrib.admin import helpers
from django.template.response import TemplateResponse
from django.utils.html import format_html

from .controllers import rerun_logged_task
from .models import LoggedTask


class LoggedTaskAdmin(admin.ModelAdmin):

    list_display = (
        "id",
        "queue",
        "task_name",
        "created_at",
        "args",
        "kwargs",
        "status",
    )
    list_filter = ("task_name", "queue", "status")

    fields = readonly_fields = (
        "id",
        "queue",
        "created_at",
        "task_name",
        "admin_args",
        "admin_kwargs",
        "admin_stacktrace",
        "status",
        "admin_rollbar",
    )
    ordering = ("-created_at",)

    actions = ["retry_logged_tasks"]

    def retry_logged_tasks(self, request, queryset):
        opts = self.model._meta

        if request.method == "POST":
            for logged_task in queryset:
                rerun_logged_task(logged_task)
            count = queryset.count()
            message = "Retried {} Logged Task items".format(count)
            self.message_user(request, message, messages.SUCCESS)
            # Return None to display the change list page again.
            return None

        context = {
            "title": "Retry Tasks?",
            "queryset": queryset,
            "opts": opts,
            "action_checkbox_name": helpers.ACTION_CHECKBOX_NAME,
            "media": self.media,
        }

        return TemplateResponse(
            request,
            "admin/retry_task_confirmation.html",
            context=context,
            current_app=self.admin_site.name,
        )

    retry_logged_tasks.short_description = "Retry Selected Tasks"

    def admin_stacktrace(self, logged_task):
        return format_html("<br/><pre>{}</pre>", logged_task.stacktrace)

    def admin_args(self, logged_task):
        args = json.loads(logged_task.args)
        return format_html("<br/><pre>{}</pre>", json.dumps(args, indent=4))

    def admin_kwargs(self, logged_task):
        kwargs = json.loads(logged_task.kwargs)
        return format_html("<br/><pre>{}</pre>", json.dumps(kwargs, indent=4))

    def admin_rollbar(self, logged_task):
        if not logged_task.rollbar_occurrence:
            return "No Rollbar occurrence logged"
        return format_html(
            "<a href='https://rollbar.com/occurrence/uuid/?uuid={}'>Rollbar occurrence</a>",
            logged_task.rollbar_occurrence,
        )


admin.site.register(LoggedTask, LoggedTaskAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-09-20 21:28
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="LoggedTask",
            fields=[
                (
                    "id",
                    models.CharField(
                        max_length=255, primary_key=True, serialize=False, unique=True
                    ),
                ),
                ("task_name", models.CharField(max_length=2056)),
                ("status", models.CharField(max_length=50)),
                ("args", models.TextField(default="[]")),
                ("kwargs", models.TextField(default="{}")),
                ("stacktrace", models.TextField(blank=True)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
            ],
        )
    ]



from mbq import metrics

from .__version__ import (  # noqa
    __author__,
    __author_email__,
    __description__,
    __license__,
    __title__,
    __url__,
    __version__,
)

from . import killswitch  # noqa

_collector: metrics.Collector

default_app_config = "mbq.ranch.apps.RanchConfig"



import logging
from typing import Callable

from django.conf import settings

import celery


logger = logging.getLogger(__name__)


def create_task_class(variation: Callable[[str, bool], bool]) -> celery.Task:
    class KillSwitchTask(celery.Task):
        def killswitch_name(self):
            service = settings.RANCH["service"]
            return f"task-killswitch-{service}-{self.name}".replace(".", "-").replace(
                "_", "-"
            )

        def __call__(self, *args, **kwargs):
            if variation(self.killswitch_name(), False):
                from . import _collector

                logger.info(f"Skipping task {self.name} with killswitch feature flag")
                _collector.increment("killswitch_on", tags={"task": self.name})
                return

            super().__call__(*args, **kwargs)

    return KillSwitchTask



# Generated by Django 2.1.7 on 2019-02-15 21:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("ranch", "0002_loggedtask_queue")]

    operations = [
        migrations.AddField(
            model_name="loggedtask",
            name="rollbar_occurrence",
            field=models.UUIDField(null=True),
        )
    ]



from django.conf import settings

from celery.signals import before_task_publish, task_prerun

from ..lib.error_handling import log_errors_and_send_to_rollbar


CORRELATION_SETTINGS = settings.RANCH.get("correlation", {})
CORRELATION_ID_GETTER = CORRELATION_SETTINGS.get("getter", lambda: None)
CORRELATION_ID_SETTER = CORRELATION_SETTINGS.get("setter", lambda value: None)


@before_task_publish.connect
@log_errors_and_send_to_rollbar
def add_correlation_id_to_headers(
    sender,
    body,
    exchange,
    routing_key,
    headers,
    properties,
    declare,
    retry_policy,
    **kwargs
):
    correlation_id = CORRELATION_ID_GETTER()
    if correlation_id:
        headers["__RANCH_CORRELATION_ID"] = correlation_id


@task_prerun.connect
@log_errors_and_send_to_rollbar
def set_correlation_id_from_headers(sender, task_id, task, **kwargs):
    correlation_id = getattr(task.request, "__RANCH_CORRELATION_ID", None)
    if correlation_id:
        CORRELATION_ID_SETTER(correlation_id)



import logging

import requests
from six.moves.urllib.parse import urlparse

from .. import _collector


logger = logging.getLogger(__name__)


def send_rabbitmq_queue_stats(broker_url, queue_names):
    queue_names = set(queue_names)
    parsed = urlparse(broker_url)

    url = "https://{}/api/queues".format(parsed.hostname)
    username = parsed.username or "guest"
    password = parsed.password or "guest"

    try:
        response = requests.get(url, auth=(username, password), timeout=0.1)
    except requests.exceptions.RequestException:
        # We set a very aggressive timeout on the request so that we do not
        # block our worker for very long. It's expected that this will
        # sometimes fail, but it's unlikely to matter since we should be
        # collecting queue stats frequently. As such, we log but otherwise
        # swallow the exception (so we do not Rollbar or emit generic error
        # metrics)
        logger.info("Error requesting metrics from RabbitMQ, skipping")
        _collector.increment("rabbitmq.stats.timeout", value=1)
        return

    response.raise_for_status()

    payload = response.json()

    for queue in payload:
        if queue["name"] in queue_names:
            _collector.gauge(
                "queue.messages.ready",
                value=queue.get("messages_ready", 0),
                tags={"queue": queue["name"]},
            )

            _collector.gauge(
                "queue.messages.unacked",
                value=queue.get("messages_unacknowledged", 0),
                tags={"queue": queue["name"]},
            )



import os

from celery import Celery
from kombu import Queue
from mbq import ranch

from . import launch_darkly


os.environ.setdefault("DJANGO_SETTINGS_MODULE", "tests.settings")
celery_app = Celery(
    "ranch_test", task_cls=ranch.killswitch.create_task_class(launch_darkly.variation)
)

celery_app.conf.update(
    broker_url="amqp://rabbitmq",
    task_default_exchange="default",
    task_default_queue="default",
    task_default_routing_key="default",
    task_soft_time_limit=600,  # raises SoftTimeLimitExceeded after 10 min
    timezone="America/New_York",
    worker_hijack_root_logger=False,
    task_queues=(Queue("default"),),
)

celery_app.autodiscover_tasks()



from mbq import env, metrics


SECRET_KEY = "fake-key"
DEBUG = True
RANCH = {"env": "Test", "service": "test-service"}
DATABASES = {"default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"}}
INSTALLED_APPS = ["mbq.ranch"]
USE_TZ = True
metrics.init("ranch-test", env.Environment.LOCAL)



import threading
import time
import uuid
from collections import Counter

from django.conf import settings

import arrow
import billiard.compat
from celery.signals import (
    after_task_publish,
    before_task_publish,
    celeryd_after_setup,
    heartbeat_sent,
    task_failure,
    task_postrun,
    task_prerun,
    task_retry,
    task_success,
    task_unknown,
    worker_process_init,
)

from .. import _collector
from ..lib.error_handling import log_errors_and_send_to_rollbar
from ..models import LoggedTask
from ..stats.rabbitmq import send_rabbitmq_queue_stats


local = threading.local()

HEARTBEAT_METRICS_INTERVAL_SECONDS = 30
EXTRA_ERROR_QUEUE_TAGS_FN = settings.RANCH.get(
    "extra_error_queue_tags_fn", lambda task: []
)


def _mem_rss_bytes():
    """
    `billiard.compat.mem_rss()` returns the memory in kB that the worker
    is using. Celery uses `mem_rss()` to enforce worker memory limits;
    Ranch uses the same function to track worker RSS over time and RSS
    change after each task is run. We convert to bytes (`* 1000`) to
    make the metrics easier to understand.
    """
    return billiard.compat.mem_rss() * 1000


@before_task_publish.connect
@log_errors_and_send_to_rollbar
def add_metadata_to_task_headers(
    sender,
    body,
    exchange,
    routing_key,
    headers,
    properties,
    declare,
    retry_policy,
    **kwargs
):
    queued_at = arrow.get(headers.get("eta")).timestamp
    headers["__RANCH_QUEUED_AT"] = queued_at


@after_task_publish.connect
@log_errors_and_send_to_rollbar
def send_metrics_on_publish(sender, headers, body, exchange, routing_key, **kwargs):
    _collector.increment(
        "task.publish", value=1, tags={"task": headers.get("task", "unknown")}
    )


@task_prerun.connect
@log_errors_and_send_to_rollbar
def store_metadata_before_task_runs(
    sender, task_id, task, args, kwargs, **extra_kwargs
):
    local.task_received_at = time.time()
    try:
        local.task_queue = task.request.delivery_info.get("routing_key")
    except Exception:
        local.task_queue = None

    local.rss_before_bytes = _mem_rss_bytes()

    queued_at = getattr(task.request, "__RANCH_QUEUED_AT", None)
    if queued_at:
        wait_time_ms = (time.time() - queued_at) * 1000
        _collector.timing(
            "task.wait_time",
            value=wait_time_ms,
            tags={"task": task.name, "queue": local.task_queue},
        )


@task_postrun.connect
@log_errors_and_send_to_rollbar
def send_metrics_after_task_runs(
    sender, task_id, task, args, kwargs, retval, state, **extra_kwargs
):
    execution_time_ms = (time.time() - local.task_received_at) * 1000
    _collector.timing(
        "task.execution_time",
        value=execution_time_ms,
        tags={"task": task.name, "state": state, "queue": local.task_queue},
    )

    local.tasks_processed += 1
    _collector.gauge(
        "worker.tasks_processed",
        value=local.tasks_processed,
        tags={"queue": local.task_queue, "worker_id": local.worker_id},
    )

    current_rss_bytes = _mem_rss_bytes()

    _collector.gauge(
        "worker.rss_bytes",
        value=current_rss_bytes,
        tags={"queue": local.task_queue, "worker_id": local.worker_id},
    )

    rss_change_bytes = current_rss_bytes - local.rss_before_bytes
    _collector.increment(
        "task.rss_change_bytes",
        value=rss_change_bytes,
        tags={"task": task.name, "queue": local.task_queue},
    )


def _send_task_processed_metric(task_name, result):
    _collector.increment(
        "task.processed",
        value=1,
        tags={"task": task_name, "result": result, "queue": local.task_queue},
    )


@task_retry.connect
@log_errors_and_send_to_rollbar
def send_metrics_on_task_retry(sender, request, reason, einfo, **kwargs):
    _send_task_processed_metric(sender.name, "retry")


@task_success.connect
@log_errors_and_send_to_rollbar
def send_metrics_on_task_success(sender, result, *args, **kwargs):
    _send_task_processed_metric(sender.name, "success")


@task_failure.connect
@log_errors_and_send_to_rollbar
def send_metrics_on_task_failure(
    sender, task_id, exception, args, kwargs, traceback, einfo, **extra_kwargs
):
    _send_task_processed_metric(sender.name, "failure")


@task_unknown.connect
@log_errors_and_send_to_rollbar
def send_metrics_on_task_unknown(sender, name, id, message, exc, **kwargs):
    _collector.increment("task.unknown", value=1, tags={"task": name})


@celeryd_after_setup.connect
@log_errors_and_send_to_rollbar
def store_metadata_on_worker_parent_start(sender, instance, conf, **kwargs):
    local.worker_instance = instance
    local.app_config = conf

    # worker_queues is a dict of queue name to Celery queue instance, for
    # all queues this worker is configured to process
    # Don't freak out about the variable named amqp, yes it is
    # called this even if you're using Redis. Celery is jank.
    local.worker_queues = instance.app.amqp.queues

    # broker_type is a string, "amqp" or "redis" (or others but these
    # are the ones we expect to use)
    local.broker_type = instance.app.broker_connection().transport_cls

    local.last_send_on_heartbeat = 0


@worker_process_init.connect
@log_errors_and_send_to_rollbar
def store_metadata_on_worker_child_start(sender, **kwargs):
    # number of tasks processed over the lifetime of each worker *child* process
    local.tasks_processed = 0
    local.worker_id = str(uuid.uuid4())


def _send_queue_length_metrics():
    if local.broker_type == "amqp":
        send_rabbitmq_queue_stats(
            local.app_config["broker_url"], list(local.worker_queues.keys())
        )


def _error_queue_tags(task):
    tags = ["queue:{}".format(task.queue), "task:{}".format(task.task_name)]
    tags.extend(EXTRA_ERROR_QUEUE_TAGS_FN(task))
    return tuple(tags)


def _send_error_queue_metrics():
    tasks = LoggedTask.objects.all()
    counts = Counter(_error_queue_tags(task) for task in tasks)

    for tags, count in counts.items():
        _collector.gauge("queue.errors.length", value=count, tags=tags)


@heartbeat_sent.connect
@log_errors_and_send_to_rollbar
def send_queue_metrics_on_heartbeat(sender, **kwargs):
    if time.time() - HEARTBEAT_METRICS_INTERVAL_SECONDS < local.last_send_on_heartbeat:
        return

    _send_queue_length_metrics()
    _send_error_queue_metrics()

    local.last_send_on_heartbeat = time.time()



import json
import uuid
from unittest.mock import Mock, patch

from django.db.models import signals
from django.test import TestCase

from mbq.ranch.constants import TaskStatus
from mbq.ranch.controllers import persist_task, rerun_logged_task
from mbq.ranch.exceptions import TaskNotFound
from mbq.ranch.models import LoggedTask


class TaskManagementTests(TestCase):
    def setUp(self):
        signals.post_save.disconnect(sender=LoggedTask, dispatch_uid="logged_task_save")
        signals.post_delete.disconnect(
            sender=LoggedTask, dispatch_uid="logged_task_delete"
        )

    def test_persist_failed_task(self):
        LoggedTask.objects.all().delete()

        persist_task(
            "task_id",
            "queue_name",
            "task_name",
            [1, 2, 3],
            {"foo": 1, "bar": 2},
            "stacktrace",
            TaskStatus.FAILURE,
            uuid.uuid4(),
        )

        logged_task = LoggedTask.objects.filter(
            id="task_id",
            task_name="task_name",
            stacktrace="stacktrace",
            status=TaskStatus.FAILURE,
        ).first()

        self.assertIsNotNone(logged_task)
        self.assertEqual(json.loads(logged_task.args), [1, 2, 3])
        self.assertEqual(json.loads(logged_task.kwargs), {"foo": 1, "bar": 2})

    @patch("mbq.ranch.controllers.import_module")
    def test_rerun_task_with_params(self, module_import_mock):
        module_mock = Mock()
        module_import_mock.return_value = module_mock

        logged_task = persist_task(
            "task_id",
            "queue_name",
            "some.module.lookup.and.thefunction",
            ["a", "b"],
            {"fizz": "buzz"},
            "stacktrace",
            TaskStatus.FAILURE,
            uuid.uuid4(),
        )
        rerun_logged_task(logged_task)

        module_import_mock.assert_called_once_with("some.module.lookup.and")
        module_mock.thefunction.delay.assert_called_once_with("a", "b", fizz="buzz")
        self.assertFalse(LoggedTask.objects.filter(pk=logged_task.pk).exists())

    @patch("mbq.ranch.controllers.import_module")
    def test_rerun_task_with_no_params(self, module_import_mock):
        module_mock = Mock()
        module_import_mock.return_value = module_mock

        logged_task = persist_task(
            "task_id",
            "queue_name",
            "some.module.lookup.and.thefunction",
            [],
            {},
            "stacktrace",
            TaskStatus.FAILURE,
            uuid.uuid4(),
        )
        rerun_logged_task(logged_task)

        module_import_mock.assert_called_once_with("some.module.lookup.and")
        module_mock.thefunction.delay.assert_called_once_with()

        self.assertFalse(LoggedTask.objects.filter(pk=logged_task.pk).exists())

    @patch("mbq.ranch.controllers.import_module")
    def test_rerun_task_not_found(self, module_import_mock):
        module_mock = Mock()
        module_import_mock.return_value = module_mock

        del module_mock.somefunction

        logged_task = persist_task(
            "task_id",
            "queue_name",
            "somemodule.somefunction",
            ["a", "b"],
            {"fizz": "buzz"},
            "stacktrace",
            TaskStatus.FAILURE,
            uuid.uuid4(),
        )

        with self.assertRaises(TaskNotFound):
            rerun_logged_task(logged_task)



import json
from importlib import import_module

from .exceptions import TaskNotFound
from .models import LoggedTask


def persist_task(
    task_id, queue, task_name, args, kwargs, stacktrace, status, rollbar_uuid
):
    return LoggedTask.objects.create(
        id=task_id,
        queue=queue,
        task_name=task_name,
        args=json.dumps(args),
        kwargs=json.dumps(kwargs),
        stacktrace=stacktrace,
        status=status,
        rollbar_occurrence=rollbar_uuid,
    )


def rerun_logged_task(logged_task):
    task_name = logged_task.task_name.split(".")[-1]
    module_path = logged_task.task_name.replace(".{}".format(task_name), "")

    module = import_module(module_path)
    task_func = getattr(module, task_name, None)

    if task_func is None:
        raise TaskNotFound()

    args = json.loads(logged_task.args)
    kwargs = json.loads(logged_task.kwargs)
    task_func.delay(*args, **kwargs)
    logged_task.delete()



import logging

from .celery import celery_app


@celery_app.task()
def hi_task():
    logging.info(">>>>>>  OOOH HAAAIII")


@celery_app.task()
def fail_task():
    logging.info(">>>>>>  FAIL TASK")
    raise ValueError



from django.db import models


class LoggedTask(models.Model):
    id = models.CharField(max_length=255, primary_key=True, unique=True)
    queue = models.CharField(max_length=128, null=True)
    task_name = models.CharField(max_length=2056)
    status = models.CharField(max_length=50)
    args = models.TextField(default="[]")
    kwargs = models.TextField(default="{}")
    stacktrace = models.TextField(blank=True)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    rollbar_occurrence = models.UUIDField(null=True)

    def __str__(self):
        return "{} - args: {}, kwargs {}".format(self.task_name, self.args, self.kwargs)



import os

import dj_database_url
from mbq import env, metrics


BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
ALLOWED_HOSTS = ["ranch.lcl.mbq.io"]
DEBUG = True
DEBUG_PROPAGATE_EXCEPTIONS = True
DATABASE_URL = "postgres://postgres:postgres@ranch-postgres:5432/ranch"
DATABASES = {
    "default": dj_database_url.parse(
        DATABASE_URL, engine="django.db.backends.postgresql", conn_max_age=0
    )
}
LAUNCHDARKLY_SDK_KEY = env.get("LAUNCHDARKLY_SDK_KEY", required=False)
SITE_ID = 1
SECRET_KEY = "BACON"
USE_I18N = True
USE_L10N = True
STATIC_URL = "/static/"
STATIC_ROOT = os.path.join(BASE_DIR, "staticfiles")
ROOT_URLCONF = "tests.urls"
TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "OPTIONS": {
            "debug": True,
            "context_processors": {
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            },
        },
    }
]
MIDDLEWARE = [
    "django.middleware.common.CommonMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
]
INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.staticfiles",
    "django.contrib.messages",
    "mbq.ranch",
    "tests",
]
PASSWORD_HASHERS = ["django.contrib.auth.hashers.MD5PasswordHasher"]
USE_TZ = True
MESSAGE_HANDLERS = "tests.message_handlers"
RANCH = {
    "env": env.Environment.LOCAL,  # e.g. production, development
    "service": "ranch",  # e.g. os-core
}
LOGGING_LEVEL = "DEBUG"
LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "verbose": {
            "format": (
                "%(levelname)-8s %(asctime)s %(name)s "
                "%(filename)s:%(lineno)s %(message)s"
            )
        }
    },
    "handlers": {
        "console": {
            "level": "INFO",
            "class": "logging.StreamHandler",
            "formatter": "verbose",
        }
    },
    "loggers": {
        "django.db.backends": {
            "handlers": ["console"],
            "level": LOGGING_LEVEL,
            "propagate": False,
        },
        "botocore": {"handlers": ["console"], "level": "ERROR", "propagate": False},
        "boto3": {"handlers": ["console"], "level": "ERROR", "propagate": False},
        "": {"handlers": ["console"], "level": LOGGING_LEVEL},
    },
}

metrics.init("ranch", env.Environment.LOCAL, {})



__author__ = "Managed by Q, Inc."
__author_email__ = "open-source@managedbyq.com"
__description__ = "MBQ Ranch"
__license__ = "Apache 2.0"
__title__ = "mbq.ranch"
__url__ = "https://github.com/managedbyq/mbq.ranch"
__version__ = "2.1.2"



import logging

from django.db.utils import Error

import rollbar
from celery.app import control

from .. import _collector


logger = logging.getLogger(__name__)


def log_errors_and_send_to_rollbar(func):
    def wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except Error:
            # If we receive any database-related exceptions, send SIGTERM--this may not be
            # necessary in every case, but it is safer
            logger.exception("Database error")
            rollbar.report_exc_info()
            control.shutdown()
        except Exception:
            logger.exception("Ranch signal error")
            rollbar.report_exc_info()
            _collector.increment("signal_error", value=1)
            raise

    return wrapper



import rollbar
from celery.signals import task_rejected

from ..lib.error_handling import log_errors_and_send_to_rollbar


@task_rejected.connect
@log_errors_and_send_to_rollbar
def handle_task_rejected(**kw):
    rollbar.report_exc_info(extra_data=kw)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-10-01 17:56
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("ranch", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="loggedtask",
            name="queue",
            field=models.CharField(max_length=128, null=True),
        )
    ]



from unittest.mock import Mock

from django.test import TestCase

from celery import Celery
from mbq import ranch


class TaskManagementTests(TestCase):
    def test_killswitch_on(self):
        celery_app = Celery(
            "ranch_test", task_cls=ranch.killswitch.create_task_class(lambda x, y: True)
        )
        mock_task_func = Mock()

        @celery_app.task()
        def test_task():
            mock_task_func()

        test_task()
        mock_task_func.assert_not_called()

    def test_killswitch_off(self):
        celery_app = Celery(
            "ranch_test",
            task_cls=ranch.killswitch.create_task_class(lambda x, y: False),
        )
        mock_task_func = Mock()

        @celery_app.task()
        def test_task():
            mock_task_func()

        test_task()
        mock_task_func.assert_called_once_with()

    def test_killswitch_name(self):
        celery_app = Celery(
            "ranch_test",
            task_cls=ranch.killswitch.create_task_class(lambda x, y: False),
        )

        @celery_app.task()
        def test_task():
            pass

        self.assertEqual(
            test_task.killswitch_name(),
            "task-killswitch-test-service-tests-tests-test-killswitch-test-task",
        )



from django.conf import settings

import ldclient


_ld_client: ldclient.LDClient
_user: dict


def init():
    global _ld_client
    global _user
    if _ld_client:
        return
    ldclient.set_config(ldclient.Config(sdk_key=settings.LAUNCHDARKLY_SDK_KEY))
    _ld_client = ldclient.get()
    _user = {"custom": {}, "key": "system", "anonymous": True}


def variation(flag_key: str, default: bool) -> bool:
    init()
    return _ld_client.variation(flag_key, _user, default)



from django.apps import AppConfig
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from mbq import metrics


class RanchConfig(AppConfig):
    name = "mbq.ranch"
    verbose_name = "Ranch"

    def ready(self):
        env = settings.RANCH.get("env")
        service = settings.RANCH.get("service")

        if not all([env, service]):
            raise ImproperlyConfigured(
                "mbq.ranch must be configured with env and service parameters."
            )

        self.module._collector = metrics.Collector(
            namespace="mbq.ranch", tags={"env": env, "service": service}
        )

        from . import signal_handlers  # noqa



import codecs
import os

import setuptools


here = os.path.abspath(os.path.dirname(__file__))

with codecs.open("README.md", "r", "utf-8") as f:
    readme = f.read()

about = {}
with codecs.open(
    os.path.join(here, "mbq", "ranch", "__version__.py"), "r", "utf-8"
) as f:
    exec(f.read(), about)

setuptools.setup(
    name=about["__title__"],
    description=about["__description__"],
    long_description=readme,
    long_description_content_type="text/markdown",
    version=about["__version__"],
    license=about["__license__"],
    url=about["__url__"],
    author=about["__author__"],
    author_email=about["__author_email__"],
    maintainer=about["__author__"],
    maintainer_email=about["__author_email__"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development :: Libraries",
    ],
    keywords="",
    package_data={"mbq.ranch": ["py.typed"]},
    packages=setuptools.find_packages(),
    install_requires=[
        "arrow",
        "celery<5.0",
        "ldclient-py>=6,<7",
        "mbq.env>=2,<3",
        "mbq.metrics>=1,<2",
        "rollbar",
    ],
    zip_safe=False,
)



"""Handlers for enforcing opinionated config. Gross, but necessary."""
import copy
import logging

from django.conf import settings

from celery.signals import celeryd_init, setup_logging

from ..lib.error_handling import log_errors_and_send_to_rollbar


@celeryd_init.connect
@log_errors_and_send_to_rollbar
def enforce_app_config(sender, instance, conf, options, **kwargs):
    # acks_late is a global config setting *but* the setting is only used to determine
    # instance properties on the Tasks when they are instantiated. I couldn't find a
    # signal that'd let me alter config before the Tasks were instantiated, hence
    # the manual task modification below. We alter the config just to keep things
    # looking relatively consistent if someone comes spelunking in here later.
    conf.task_acks_late = True

    for task in instance.app.tasks.values():
        task.acks_late = True


@setup_logging.connect
@log_errors_and_send_to_rollbar
def setup_logging(loglevel, logfile, format, colorize, **kwargs):
    config = copy.copy(settings.LOGGING)
    config["disable_existing_loggers"] = False
    logging.config.dictConfig(config)
    logging.info("logging set up")



import traceback as traceback_stdlib

import rollbar
from celery.signals import task_failure, task_unknown

from ..constants import TaskStatus
from ..controllers import persist_task
from ..lib.error_handling import log_errors_and_send_to_rollbar


@task_failure.connect
@log_errors_and_send_to_rollbar
def handle_task_failure(
    sender, task_id, exception, args, kwargs, traceback, einfo, **extra_kwargs
):
    rollbar_uuid = rollbar.report_exc_info(extra_data=extra_kwargs)
    try:
        persist_task(
            task_id,
            sender.request.delivery_info.get("routing_key"),
            sender.name,
            args,
            kwargs,
            traceback_stdlib.format_exc(),
            TaskStatus.FAILURE,
            rollbar_uuid,
        )
    except Exception:
        # If anything goes wrong persisting the task to the DB,
        # we put it back in the Celery queue so that it is not
        # permanently lost. If that also fails, we should be
        # saved by acks_late=True since the original job
        # can only be acknowledged if we can still talk to Rabbit.
        sender.retry(countdown=60)


@task_unknown.connect
@log_errors_and_send_to_rollbar
def handle_task_unknown(sender, name, id, message, exc, **kwargs):
    rollbar_uuid = rollbar.report_exc_info(extra_data=kwargs)
    try:
        queue = message.delivery_info.get("routing_key", "unknown")
    except Exception:
        queue = "unknown"

    persist_task(
        id,
        queue,
        name,
        message.payload[0],
        message.payload[1],
        traceback_stdlib.format_exc(),
        TaskStatus.UNKNOWN,
        rollbar_uuid,
    )
